"""
Reader compatibility mapping generated from test results.

This mapping shows which DatasetType readers can successfully read data
written by each (arrow_type, writer_dataset_type, content_type) combination.

Keys: (arrow_type, writer_dataset_type, content_type)
Values: List of compatible DatasetType enum values
"""

from deltacat.types.tables import DatasetType

# Mapping of (arrow_type, writer_dataset_type, content_type) -> list of compatible readers
READER_COMPATIBILITY_MAPPING = {
    ("binary", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
    ],
    ("binary", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("binary", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("bool", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date32[day]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.RAY_DATASET,
    ],
    ("date64[ms]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal128", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal256", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal256", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal256", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal256", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal256", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("decimal256", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("dictionary", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("dictionary", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("dictionary", "polars", "application/feather"): [DatasetType.POLARS],
    ("dictionary", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("dictionary", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("dictionary", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("double", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ms]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ms]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ms]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ms]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ms]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ms]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ms]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ms]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ns]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ns]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ns]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ns]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ns]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ns]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ns]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[ns]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[s]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[s]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[s]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[s]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[s]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[s]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[s]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[s]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[us]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[us]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[us]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[us]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[us]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[us]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[us]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("duration[us]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_shape_tensor", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_shape_tensor", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_shape_tensor", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_shape_tensor", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_shape_tensor", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_shape_tensor", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_shape_tensor", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_shape_tensor", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("fixed_size_list", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("float", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("halffloat", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int16", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int32", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int64", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("int8", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
    ],
    ("large_binary", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_binary", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list_view", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list_view", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list_view", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_list_view", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("large_string", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list_view", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list_view", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list_view", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("list_view", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("map", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("map", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("map", "polars", "application/feather"): [DatasetType.PANDAS, DatasetType.POLARS],
    ("map", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("map", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("map", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.RAY_DATASET,
    ],
    ("map", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("map", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("map", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("map", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("month_day_nano_interval", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("month_day_nano_interval", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("null", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("null", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.RAY_DATASET,
    ],
    ("null", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("null", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("null", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("null", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("null", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("null", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("null", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.RAY_DATASET,
    ],
    ("string", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("string", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
    ],
    ("struct", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("struct", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time32[ms]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time32[ms]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time32[ms]", "pyarrow", "application/feather"): [
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time32[ms]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time32[s]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time32[s]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time32[s]", "pyarrow", "application/feather"): [
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time32[s]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.POLARS,
        DatasetType.RAY_DATASET,
    ],
    ("time64[ns]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[ns]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[ns]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[ns]", "pyarrow", "application/feather"): [
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[ns]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[ns]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[us]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[us]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[us]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[us]", "pyarrow", "application/feather"): [
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[us]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("time64[us]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ms]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[ns]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[s]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "pandas", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "polars", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "pyarrow", "application/avro"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp[us]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ms]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[ns]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[s]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "pandas", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "polars", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "pyarrow", "application/orc"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("timestamp_tz[us]", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint16", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint16", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint16", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint16", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint16", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint16", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint16", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint16", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint32", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint32", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint32", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint32", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint32", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint32", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint32", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint32", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint64", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint64", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint64", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint64", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint64", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint64", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint64", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint64", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint8", "daft", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint8", "pandas", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint8", "pandas", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint8", "polars", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint8", "polars", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint8", "pyarrow", "application/feather"): [
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint8", "pyarrow", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
    ("uint8", "ray_dataset", "application/parquet"): [
        DatasetType.DAFT,
        DatasetType.PANDAS,
        DatasetType.POLARS,
        DatasetType.PYARROW,
        DatasetType.RAY_DATASET,
    ],
}


def get_compatible_readers(
    arrow_type: str, writer_dataset_type: str, content_type: str
):
    """Get list of compatible reader DatasetTypes for given combination."""
    key = (arrow_type, writer_dataset_type, content_type)
    compatible_readers = READER_COMPATIBILITY_MAPPING.get(key, [])
    if (
        DatasetType.PANDAS in compatible_readers
        and DatasetType.NUMPY not in compatible_readers
    ):
        compatible_readers = compatible_readers + [DatasetType.NUMPY]
    return compatible_readers


def is_reader_compatible(
    arrow_type: str,
    writer_dataset_type: str,
    content_type: str,
    reader_dataset_type: DatasetType,
) -> bool:
    """Check if a specific reader is compatible with given combination."""
    compatible_readers = get_compatible_readers(
        arrow_type, writer_dataset_type, content_type
    )
    return reader_dataset_type in compatible_readers
